﻿-------------------------------------------------------------------------------
-- local cache

local CreateFrame = CreateFrame;

nUI.Class.StatusBar = nUI.Class:CreatePrototype();
local prototype     = nUI.Class.StatusBar.prototype;


function prototype:init( parent, id, layout )

	self.parent      = parent;
	self.id          = id;
	self.rid         = 1;
	self.initialized = false;
	self.layout      = layout;
	self.cur_length  = nil;
	self.offset      = nil;
	self.bar_width   = layout.width;
	self.elapsed     = 0;
	self.x1          = 0;
	self.x2          = 1;
	
	self:layoutFrame();

	if self.layout.Colors.animation ~= "" then
		self.frame.SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	end
	
end

-------------------------------------------------------------------------------

function prototype:IsShown()
	
	return self.frame:IsShown();
	
end

-------------------------------------------------------------------------------

function prototype:SetScale( scale )
	
	self.frame:SetScale( scale );
	
end

-------------------------------------------------------------------------------

function prototype:Hide()
	
	self.frame:Hide();
	self.frame:SetScript( "OnUpdate", nil );
	self.showing = false;
	
end

-------------------------------------------------------------------------------

function prototype:Show()
	
	self.elapsed = 0;
	self.showing = true;

	self.frame:Show();

	if self.layout.Colors.animation ~= "" then
		self.frame.SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	end
		
end

-------------------------------------------------------------------------------

function prototype:layoutFrame()

	self.frame       = CreateFrame( "Frame", self.layout.name, self.parent );
	self.frame.class = self;

	local color = self.layout.Colors.bg;
	
	self.frame:SetWidth( self.layout.width );
	self.frame:SetHeight( self.layout.height );
	self.frame.texture = self.frame:CreateTexture();
	self.frame.texture:SetAllPoints( self.frame );
	self.frame.texture:SetTexture( color.r, color.g, color.b, color.a );
	
	if self.layout.Colors.animation ~= "" then
		self.frame.anim = CreateFrame( "Frame", self.layout.name.."_Anim", self.frame );
		self.frame.anim:SetAllPoints( self.frame );

		if self.layout.Options.grow ~= "left" then
			self.frame.anim:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 0, 0 );
			self.frame.anim:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 0 );
			self.frame.anim:SetPoint( "TOPRIGHT", self.frame, "TOPLEFT", self.bar_width, 0 );
			self.frame.anim:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMLEFT", self.bar_width, 0 );
		else
			self.frame.anim:SetPoint( "TOPRIGHT", self.frame, "TOPRIGHT", 0, 0 );
			self.frame.anim:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", 0, 0 );
			self.frame.anim:SetPoint( "TOPLEFT", self.frame, "TOPRIGHT", -self.bar_width, 0 );
			self.frame.anim:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMRIGHT", -self.bar_width, 0 );
		end
		
		self.frame.anim.texture = self.frame.anim:CreateTexture();
		self.frame.anim.texture:SetTexture( self.layout.Colors.animation );
		self.frame.anim.SetAllPoints( self.frame.anim );
		self.frame.anim.texture:SetTexCoordModifiesRect( true );
	end
	
	self.frame.bar = CreateFrame( "Frame", self.layout.name.."_Bar", self.frame );
	
	if self.frame.anim then
		self.frame.bar.SetAllPoints( self.frame.anim );
		self.frame.bar:SetFrameLevel( self.frame.animr:GetFrameLevel()+1 );
	elseif self.layout.Options.grow ~= "left" then
		self.frame.bar:SetPoint( "TOPLEFT", self.frame, "TOPLEFT", 0, 0 );
		self.frame.bar:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMLEFT", 0, 0 );
		self.frame.bar:SetPoint( "TOPRIGHT", self.frame, "TOPLEFT", self.bar_width, 0 );
		self.frame.bar:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMLEFT", self.bar_width, 0 );
	else
		self.frame.bar:SetPoint( "TOPRIGHT", self.frame, "TOPRIGHT", 0, 0 );
		self.frame.bar:SetPoint( "BOTTOMRIGHT", self.frame, "BOTTOMRIGHT", 0, 0 );
		self.frame.bar:SetPoint( "TOPLEFT", self.frame, "TOPRIGHT", -self.bar_width, 0 );
		self.frame.bar:SetPoint( "BOTTOMLEFT", self.frame, "BOTTOMRIGHT", -self.bar_width, 0 );
	end
	
	self.frame.bar.center = CreateFrame( "Frame", self.layout.name.."_Center", self.frame.bar );
	self.frame.bar.center:SetAllPoints( self.frame.bar );
	
	self.frame.bar.center.text = self.frame.bar.center:CreateFontString( self.layout.name.."_CenterText", "OVERLAY" );
	self.frame.bar.center.text:SetWidth( self.layout.width );
	self.frame.bar.center.text:SetHeight( self.layout.height );
	self.frame.bar.center.text:SetPoint( "CENTER", self.frame.bar.center, "CENTER", 0, 1 * nUI.scale );
	self.frame.bar.center.text:SetFont( "Fonts/ARIALN.TTF", self.layout.height * 0.7, "OUTLINE" );
	self.frame.bar.center.text:SetTextColor( 1, 1, 1, 1 );
	self.frame.bar.center.text:SetJustifyH( "CENTER" );
	self.frame.bar.center.text:SetJustifyV( "MIDDLE" );
	
	self.frame.bar.left = CreateFrame( "Frame", self.layout.name.."_Left", self.frame.bar );
	self.frame.bar.left:SetAllPoints( self.frame.bar );
	
	self.frame.bar.left.text = self.frame.bar.left:CreateFontString( self.layout.name.."_LeftText", "OVERLAY" );
	self.frame.bar.left.text:SetWidth( self.layout.width );
	self.frame.bar.left.text:SetHeight( self.layout.height );
	self.frame.bar.left.text:SetPoint( "CENTER", self.frame.bar.center, "CENTER", 0, 1 * nUI.scale );
	self.frame.bar.left.text:SetFont( "Fonts/ARIALN.TTF", self.layout.height * 0.7, "OUTLINE" );
	self.frame.bar.left.text:SetTextColor( 1, 1, 1, 1 );
	self.frame.bar.left.text:SetJustifyH( "LEFT" );
	self.frame.bar.left.text:SetJustifyV( "MIDDLE" );
	
	self.frame.bar.right = CreateFrame( "Frame", self.layout.name.."_Right", self.frame.bar );
	self.frame.bar.right:SetAllPoints( self.frame.bar );
	
	self.frame.bar.right.text = self.frame.bar.right:CreateFontString( self.layout.name.."_RightText", "OVERLAY" );
	self.frame.bar.right.text:SetWidth( self.layout.width );
	self.frame.bar.right.text:SetHeight( self.layout.height );
	self.frame.bar.right.text:SetPoint( "CENTER", self.frame.bar.center, "CENTER", 0, 1 * nUI.scale );
	self.frame.bar.right.text:SetFont( "Fonts/ARIALN.TTF", self.layout.height * 0.7, "OUTLINE" );
	self.frame.bar.right.text:SetTextColor( 1, 1, 1, 1 );
	self.frame.bar.right.text:SetJustifyH( "RIGHT" );
	self.frame.bar.right.text:SetJustifyV( "MIDDLE" );
	
	self.frame.bar.texture = self.frame.bar:CreateTexture();
	self.frame.bar.texture:SetTexCoordModifiesRect( true );
	self.frame.bar.texture:SetAllPoints( self.frame.bar );

	if self.layout.Options.grow == "left" then
		self.frame.bar.texture:SetPoint( "RIGHT", self.frame.bar, "RIGHT", 0, 0 );
	else
		self.frame.bar.texture:SetPoint( "LEFT", self.frame.bar, "LEFT", 0, 0 );
	end

	if self.layout.Colors.overlay ~= nil then
		self.frame.overlay = CreateFrame( "Frame", self.layout.name.."_Overlay", self.frame );
		self.frame.overlay:SetAllPoints( self.frame.bar );
		self.frame.overlay:SetFrameLevel( self.frame.bar:GetFrameLevel()+1 );

		self.frame.overlay.texture = self.frame.overlay:CreateTexture();
		self.frame.overlay.texture:SetTexture( self.layout.Colors.overlay );
		self.frame.overlay.texture:SetAllPoints( self.frame.overlay );
		self.frame.overlay.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
	end
	
	if self.layout.Options.wrap ~= "" then	
		self.frame.wrap = CreateFrame( "Frame", self.layout.name.."_Wrap", self.frame );
		self.frame.wrap:SetAllPoints( self.frame.bar );
		
		self.frame.wrap.texture = self.frame.wrap:CreateTexture();
		self.frame.wrap.texture:SetTexture( self.layout.Options.wrap );
		self.frame.wrap.texture:SetPoint( "TOP", self.frame.wrap, "TOPLEFT", 0, 0 );
		self.frame.wrap.texture:SetPoint( "BOTTOM", self.frame.wrap, "BOTTOMLEFT", 0, 0 );
	end
end

-------------------------------------------------------------------------------

function prototype:SetPoint( point, relative_to, relative_point, ofsX, ofsY )

	self.frame:SetPoint( point, relative_to, relative_point, ofsX, ofsY );
	
end

function prototype:SetAllPoints( relative_to )
	
	self.frame:SetAllPoints( relative_to );
	
end

-------------------------------------------------------------------------------

function prototype:updateValues( cur_val, offset, min_val, max_val )
	
	local	redraw = false;
	local	recolor;
	local	bar_ratio;
	local	bar_length;
	local	offset_ratio;
	local	offset_len;

--	nUI:debugMsg( self.layout.name.." updating values: current="..cur_val..", offset="..offset..", min_val="..min_val..", max_val="..max_val );

	if offset  == nil then offset  = self.layout.Values.offset; end
	if min_val == nil then min_val = self.layout.Values.min; end
	if max_val == nil then max_val = self.layout.Values.max; end
	
	-- sanity check the data so we don't blow up the bar
	
	if max_val < min_val then
		nUI:debugMsg( 1, "max value is less than min ["..max_val.." < "..min_val.."] in StatusBar ["..(self.layout.name).."]" );
		max_val = min_val;
	end
	
	if cur_val < min_val then
		nUI:debugMsg( 1, "current value is less than min ["..cur_val.." < "..min_val.."] in StatusBar ["..(self.layout.name).."]" );
		cur_val = min_val;
	elseif cur_val > max_val then
		nUI:debugMsg( 1, "current value is greater than max ["..cur_val.." > "..max_val.."] in StatusBar ["..(self.layout.name).."]" );
		cur_val = max_val;
	end
	
	if offset < min_val then
		nUI:debugMsg( 1, "offset value is less than min ["..offset.." < "..min_val.."] in StatusBar ["..(self.layout.name).."]" );
		offset = min_val;
	elseif offset > max_val then
		nUI:debugMsg( 1, "offset value is greater than max ["..offset.." > "..max_val.."] in StatusBar ["..(self.layout.name).."]" );
		offset = max_val;
	end
	
	-- this test is, technically, an OR operation and could be done on a single
	-- line. However, rather than rely on the interpreter optimizing, we make sure
	-- the test is optimized by breaking on the first true condition.
	-- if the frame isn't visible, we don't do anything regardless of changes
	
	if self.layout.Values.cur ~= cur_val       then redraw = true;
	elseif self.layout.Values.max ~= max_val   then redraw = true;
	elseif self.layout.Values.min ~= min_val   then redraw = true;
	elseif self.layout.Values.offset ~= offset then redraw = true;
	elseif self.cur_length == nil              then redraw = true;
	else redraw = false;
	end
	
	-- we don't do anything unless a value changed (be nice the CPU please)
	
	if redraw then

		-- remember our current state
		
		self.layout.Values.min    = min_val;
		self.layout.Values.max    = max_val;
		self.layout.Values.cur    = cur_val;
		self.layout.Values.offset = offset;
		
		-- calculate the current value in the range 0 to 1 relative to the min and max
	
		if min_val == max_val then
			
			bar_ratio    = 0;
			bar_length   = 0;
			offset_ratio = 0;
			offset_len   = 0;
			
		else
			
			bar_ratio    = (cur_val - min_val) / (max_val - min_val);
			bar_length   = math.floor( self.bar_width * bar_ratio );
			offset_ratio = (offset - min_val) / (max_val - min_val);
			offset_len   = math.floor( self.bar_width * offset_ratio );
			
		end
		
		-- even though the current value may have changed by some amount, it does not
		-- mean that the bar length has changed in a visible amount. We discard fractional
		-- changes and only update on whole number changes in order to optimize performance
		
		if self.cur_length == bar_length and self.offset == offset_len then
			redraw = false;
		else
			self.cur_length = bar_length;
			self.offset     = offset_len;
		end
		
		-- we have to update labels even though the bar may not have changed... i.e. we 
		-- change targets that have different max values for health and mana but are
		-- both at 100% in which case the labels change but the bars don't
		
		self:updateLabels();
		
	end
		
	-- if there is a notable change in status, then we update the bar
	
	if redraw then 

		-- determine if there's a color change to be made in the bar texture
		
		local dR, dG, dB, dA, dX, r, g, b, a;
		
		if bar_ratio < 0.5 then
			r  = self.layout.Colors.min.r;
			g  = self.layout.Colors.min.g;
			b  = self.layout.Colors.min.b;
			a  = self.layout.Colors.min.a;
			dR = self.layout.Colors.mid.r - r;
			dG = self.layout.Colors.mid.g - g;
			dB = self.layout.Colors.mid.b - b;
			dA = self.layout.Colors.mid.a - a;
			dX = bar_ratio * 2;
		else
			r  = self.layout.Colors.mid.r;
			g  = self.layout.Colors.mid.g;
			b  = self.layout.Colors.mid.b;
			a  = self.layout.Colors.mid.a;
			dR = self.layout.Colors.max.r - r;
			dG = self.layout.Colors.max.g - g;
			dB = self.layout.Colors.max.b - b;
			dA = self.layout.Colors.max.a - a;
			dX = (bar_ratio - 0.5) * 2;
		end

		recolor = false;
		
		if self.color_set ~= true then
			self.color_set = true;
			recolor = true;
		end
		
		if dR ~= 0 and dX > 0 then 
			r = r + (dR * dX);
		end
		
		if dG ~= 0 and dX > 0 then 
			g = g + (dG * dX);
		end
		
		if dB ~= 0 and dX > 0 then 
			b = b + (dB * dX);
		end
		
		if dA ~= 0 and dX > 0 then 
			a = a + (dA * dX);
		end

		if recolor or self.last_r ~= r or self.last_g ~= g or self.last_b ~= b or self.last_a ~= a then
			self.last_r = r;
			self.last_g = g;
			self.last_b = b;
			self.last_a = a;
			recolor = true;
		end
		
		if recolor then
			self.frame.bar.texture:SetTexture( r, g, b, a );
		end		
		
		self:redrawBar(); 
		
	end
end

-------------------------------------------------------------------------------

function prototype:redrawBar()
	
	-- we don't *have* to use separate logic for whether we are or are not wrapping
	-- the bar. However, bar wrap uses extra math that loads the system more and is
	-- used far less often, so we test for wrap and use more efficient code if it is
	-- not enabled. 
	--
	-- NOTE: Values.offset only has meaning if Options.wrap ~= ""
	
	local x1, x2;
	
	if self.layout.Options.wrap == "" then
		
--		if self.layout.Options.grow == "left" then
--			x1 = self.bar_width - self.cur_length;
--			x2 = self.bar_width;
--		else
			x1 = 0;
			x2 = self.cur_length;
--		end
		
	else
		
		local effective_len = self.cur_length + self.offset;	
		local end_point     = min( effective_len, self.bar_width );
		local wrap          = effective_len - end_point;
		local x3;
		
--		if self.layout.Options.grow == "left" then
--			x1 = self.bar_width - end_point;
--			x2 = self.bar_width - self.offset;
--			x3 = self.bar_width - wrap;
--		else
			x1 = self.offset;
			x2 = end_point;
			x3 = wrap;
--		end
		
		-- set the position of the wraparound marker
		
		if wrap == 0 then 
			
			self.frame.wrap:Hide();
		
		else
			
			self.frame.wrap:Show();
			self.frame.wrap.texture:SetPoint( "TOP", self.frame.wrap, "TOPLEFT", x3, 0 );
			self.frame.wrap.texture:SetPoint( "BOTTOM", self.frame.wrap, "BOTTOMLEFT", x3, 0 );
			
		end
	end

	-- draw the bar and update the animation

	self.x1 = x1 / self.bar_width;
	self.x2 = x2 / self.bar_width;

	self.frame.bar.texture:SetWidth( x2-x1 );
	self.frame.bar.texture:SetHeight( self.frame:GetHeight() );

	if self.layout.Options.grow == "left" then
		self.frame.bar.texture:SetPoint( "TOPLEFT", self.frame.bar, "TOPLEFT", self.bar_width-x2+x1, 0)
		
		if self.frame.overlay then
			self.frame.overlay.texture:SetPoint( "TOPLEFT", self.frame.overlay, "TOPLEFT", self.bar_width-x2+x1, 0)
		end
		
		if self.frame.anim then
			self.frame.anim.texture:SetPoint( "TOPLEFT", self.frame.anim, "TOPLEFT", self.bar_width-x2+x1, 0)
		end
	else
		self.frame.bar.texture:SetPoint( "TOPRIGHT", self.frame.bar, "TOPRIGHT", -(self.bar_width-x2+x1), 0)
		
		if self.frame.overlay then
			self.frame.overlay.texture:SetPoint( "TOPRIGHT", self.frame.overlay, "TOPRIGHT", -(self.bar_width-x2+x1), 0)
		end
		
		if self.frame.anim then
			self.frame.anim.texture:SetPoint( "TOPRIGHT", self.frame.anim, "TOPRIGHT", -(self.bar_width-x2+x1), 0)
		end
	end

	
	if self.frame.anim then 
		self:updateAnimation();
	end		
end

-------------------------------------------------------------------------------

function prototype:updateLabels()
	
	self.lbl_min     = nil;
	self.lbl_center  = nil;
	self.lbl_max     = nil;
	self.lbl_outside = nil;
	
	if self.layout.Options.bar_name ~= "" then
		self:addLabel( self.layout.Options.show_name, self.layout.Options.bar_name );
	end
	
	if self.layout.Options.min_value ~= "none" then
		local	postfix = "";
		local	value   = self.layout.Values.min;
		
		if value >= 1000000 then
			value = nUI:round( value / 1000000, 1 );
			postfix = "M";
		elseif value >= 10000 then
			value = nUI:round( value / 1000, 1 );
			postfix = "K";
		end

		if math.floor( value ) == value and postfix ~= "" then
			self:addLabel( self.layout.Options.min_value, value..postfix );
		else
			self:addLabel( self.layout.Options.min_value, value..".0"..postfix );
		end
	end
	
	if self.layout.Options.cur_value ~= "none" then
		local	postfix = "";
		local	value   = self.layout.Values.cur;
		
		if value >= 1000000 then
			value = nUI:round( value / 1000000, 1 );
			postfix = "M";
		elseif value >= 10000 then
			value = nUI:round( value / 1000, 1 );
			postfix = "K";
		end

		if math.floor( value ) == value and postfix ~= "" then
			self:addLabel( self.layout.Options.cur_value, value..".0"..postfix );
		else
			self:addLabel( self.layout.Options.cur_value, value..postfix );
		end
	end
	
	if self.layout.Options.max_value ~= "none" then
		local	postfix = "";
		local	value   = self.layout.Values.max;
		
		if value >= 1000000 then
			value = nUI:round( value / 1000000, 1 );
			postfix = "M";
		elseif value >= 10000 then
			value = nUI:round( value / 1000, 1 );
			postfix = "K";
		end

		if math.floor( value ) == value and postfix ~= "" then
			self:addLabel( self.layout.Options.max_value, value..".0"..postfix );
		else
			self:addLabel( self.layout.Options.max_value, value..postfix );
		end
	end
	
	if self.layout.Options.percent ~= "none" then
		local	postfix = "%";
		local	value   = self.layout.Values.cur / self.layout.Values.max;
		
		value = nUI:round( value * 100, 0 );
		
		self:addLabel( self.layout.Options.percent, value..postfix );
	end
	
	self:showLabel( "min", self.lbl_min );
	self:showLabel( "center", self.lbl_center );
	self:showLabel( "max", self.lbl_max );
	
end

-------------------------------------------------------------------------------

function prototype:addLabel( location, text )
	
	if location == "min" then
		if self.lbl_min ~= nil then self.lbl_min = self.lbl_min.."/"..text;
		else self.lbl_min = text;
		end
	elseif location == "max" then
		if self.lbl_max ~= nil then self.lbl_max = self.lbl_max.."/"..text;
		else self.lbl_max = text;
		end
	elseif location == "center" then
		if self.lbl_center ~= nil then self.lbl_center = self.lbl_center.."/"..text;
		else self.lbl_center = text;
		end
	end
end

-------------------------------------------------------------------------------

function prototype:showLabel( location, text )

	if text == nil then text = ""; end
	
	if location ~= "none" and location ~= nil then
		if location == "center" then
			self.frame.bar.center.text:SetText( text );
		elseif location == "min" then
			if self.layout.Options.grow ~= "left" then
				self.frame.bar.left.text:SetText( text );
			else
				self.frame.bar.right.text:SetText( text );
			end
		elseif location == "max" then
			if self.layout.Options.grow ~= "left" then
				self.frame.bar.right.text:SetText( text );
			else
				self.frame.bar.left.text:SetText( text );
			end
		end
	end
end

-------------------------------------------------------------------------------

function prototype:updateAnimation()
	
	if self.frame.isVisible() then
	end
	
end

-------------------------------------------------------------------------------

function prototype:onUpdate( elapsed )
	
	if self.layout.Colors.animation ~= "" then
		
		self.elapsed = self.elapsed + elapsed;
		
		if self.elapsed > 0.08 then -- update animation 12.5 times per second
		
			self.elapsed = 0;
			self:updateAnimation();
			
		end
	end
end
